// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_multifit_nlin.h>

static double fft_local_T=300.0;
double mull=1e11;
static char global_output[1000];
static int extra=10;
int fft_len=100; // was 30
int fit_stop=35+10;//+20+40;
int fit_start=5;

int dolog=FALSE;
static int donotdump=TRUE;
struct data {
size_t n;
double * y;
double * sigma;
};

struct istruct main_sigma;

struct istruct main_g;
struct istruct main_exp;
struct istruct main_error;
struct istruct main_fit;
struct istruct main_g_save;
struct istruct main_simple;
struct istruct omega_scale;

static double v0=2e7;
static double sigma=1e-25;


static double error_tot=0.0;

void g_push()
{
inter_copy(&main_g_save,&main_g,FALSE);
}

void g_pop()
{
inter_copy(&main_g,&main_g_save,FALSE);
}

double sech(double x)
{
return 2.0/(exp(x)+exp(-x));
}

void update_g(const gsl_vector * x)
{
int n;
for (n=0;n<main_g.len;n++)
{
	if (n<fit_start)
	{
		printf("a\n");
		main_g.data[n]=main_g.data[fit_start];//*exp((main_g.x[n]-main_g.x[fit_stop-1])/22e-3);//fabs(gsl_vector_get(x, 0));
	}else
	if (n<fit_stop)
	{
		printf("b\n");
		main_g.data[n]=fabs(gsl_vector_get(x, n-fit_start));
	}else
	{
		printf("c\n");
		main_g.data[n]=1.0;
	}
printf("%d %le %d\n",n,main_g.data[n],n-fit_start);
}
}

void get_fft_error ()
{


int i;
int ii;


double En=0.0;
double Ep=0.0;
double E=0.0;
double dE=0.0;
double Eomega=0.0;
double sum=0.0;

for (ii=1;ii<main_exp.len;ii++)
{

Eomega=main_exp.x[ii];
sum=0.0;
	for (i=1;i<main_g.len;i++)
	{

		if (i==1)
		{
			En=main_g.x[i];
		}else
		{
			En=main_g.x[i-1];
		}

		if (i==main_g.len-1)
		{
			Ep=main_g.x[i];
		}else
		{
			Ep=main_g.x[i+1];
		}

		E=main_g.x[i];
		dE=fabs((Ep-En)/2.0);

		//if (i==ii)
		//{
		//	printf("%le %le\n",main_exp.x[ii],main_g.x[i]);
		//	getchar();
		//}
		sum+=main_g.data[i]*(sigma*v0/2.0)*sech(q*(Eomega-E)/(fft_local_T*kb))*dE;

		//if (i==ii) sum=main_g.data[i];//*(sigma*v0/2.0)*sech(q*(Eomega-E)/(fft_local_T*kb))*dE;

		if (sum!=sum)
		{
			printf("now exit %le %le\n",main_g.data[i],(sigma*v0/2.0)*sech(q*(Eomega-E)/(fft_local_T*kb))*dE);
			exit(0);
		}
	}

main_fit.x[ii]=main_exp.x[ii];

main_fit.data[ii]=sum;


main_error.x[ii]=main_exp.x[ii];
main_error.data[ii]=fabs(main_fit.data[ii]-main_exp.data[ii]);
}
error_tot=0.0;
if (dolog==TRUE) inter_log_y(&main_fit);
//printf("Dump\n");
if (donotdump==FALSE)
{
inter_save_a(&main_exp,global_output,"rod_exp.dat");
inter_save_a(&main_fit,global_output,"rod_fit.dat");
inter_save_a(&main_error,global_output,"rod_error.dat");
inter_save_a(&main_g,global_output,"rod_g.dat");
}
return;
}


int expb_f (const gsl_vector * x, void *data, gsl_vector * f)
{
printf("expb_f\n");
double *sigma = main_sigma.data;

update_g(x);
//getchar();
get_fft_error();

size_t i;

for (i = fit_start; i < fit_stop; i++)
{
	/* Model Yi = A * exp(-lambda * i) + b */
	double Yi = main_fit.data[i];//A * exp (-lambda * t) + b;
double error=Yi - main_exp.data[i];

		gsl_vector_set (f, i-fit_start, error/sigma[i]);

	printf("test=%le\n",error/sigma[i]);
	//getchar();

}
printf("exit expb_f\n\n");
return GSL_SUCCESS;
}


int expb_df (const gsl_vector * x, void *data, gsl_matrix * J)
{
int n;
double *sigma = main_sigma.data;
printf("expb_df\n");

update_g(x);

get_fft_error();

size_t i;
int ii;
g_push();
for (i = fit_start; i <fit_stop; i++)
{
	double s = sigma[i];

	/* Jacobian matrix J(i,j) = dfi / dxj, */
	/* where fi = (Yi - yi)/sigma[i],      */
	/*       Yi = A * exp(-lambda * i) + b  */
	/* and the xj are the parameters (A,lambda,b) */
	for (ii=fit_start;ii<fit_stop;ii++)
	{
		get_fft_error ();
		double y0=0.0;

		y0=main_fit.data[i];

		double delta=fabs(main_g.data[ii]*0.001);
		main_g.data[ii]+=delta;
		get_fft_error ();
		double y1=0.0;

		y1=main_fit.data[i];

		double deriv=(y1-y0);
		deriv/=delta;

		gsl_matrix_set (J, i-fit_start, ii-fit_start, deriv/s);
		g_pop();
	}
//getchar();
printf("J = %d %d\n",i,ii);
}
printf("exit expb_df\n");
return GSL_SUCCESS;
}

int expb_fdf (const gsl_vector * x, void *data, gsl_vector * f, gsl_matrix * J)
{
printf("expb_fdf\n");
expb_f (x, data, f);
expb_df (x, data, J);
printf("exit expb_fdf\n");
//getchar();
return GSL_SUCCESS;
}

double fft_mag(struct istruct* in,int pos)
{
return sqrt(in->data[pos*2]*in->data[pos*2]+in->data[(pos*2)+1]*in->data[(pos*2)+1]);
}

void window(struct istruct* in,double left,double right)
{
int i;
struct istruct temp;
inter_new(&temp,in->len+extra+extra);

for (i=0;i<in->len;i++)
{
	temp.data[i+extra]=in->data[i];
	temp.x[i+extra]=in->x[i];
}

double delta=in->x[in->len-1]-in->x[in->len-2];
double pos=in->x[in->len-1]+delta;
for (i=0;i<extra;i++)
{
	temp.data[in->len+extra+i]=in->data[in->len-1]*right;
	temp.x[in->len+extra+i]=pos;
pos+=delta;
}
delta=fabs(in->x[2]-in->x[1]);
pos=in->x[2]-delta;
for (i=extra;i>=0;i--)
{
	temp.data[i]=0.0;
	temp.x[i]=pos;
	pos+=-delta;
}

inter_free(in);
inter_copy(in,&temp,TRUE);

inter_free(&temp);
}


void fft_extract(char *outputpath,char *infile,double T,double E_start,double E_stop,double t_start,double t_stop,int do_fit,double norm_E_pos)
{
fft_local_T=T;
int i;
double phase=0.0;
double mag=0.0;
struct istruct fft;
struct istruct fft_in;
struct istruct data;
struct istruct fft_mag;
struct istruct fft_phase;
FILE *out;
char name[1000];
strcpy(global_output,outputpath);
inter_load_a(&data,infile,0,1);
inter_chop(&data,t_start, t_stop); //1e-9 0.01
inter_save_a(&data,outputpath,"./smooth.dat");

inter_new(&fft,fft_len*2);
inter_new(&fft_in,data.len*2);

i_to_fft(&fft_in,&data);

build_dft_energy_mesh(&fft,E_start, E_stop,T);
do_dft(&fft,&fft_in);

inter_new(&fft_mag,fft_len);
fft_to_mag(&fft_mag,&fft);

inter_new(&fft_phase,fft_len);
fft_to_phase(&fft_phase,&fft);

inter_save_a(&fft_mag,"./","fft_mag.dat");
inter_save_a(&fft_phase,"./","fft_phase.dat");

inter_copy(&omega_scale,&fft_mag,TRUE);

inter_new(&main_g,fft_mag.len);
inter_new(&main_simple,data.len);
inter_new(&main_g_save,fft_mag.len);
inter_new(&main_fit,fft_mag.len);
inter_set_value(&main_fit,1e10);
inter_new(&main_error,fft_mag.len);
inter_new(&main_exp,fft_mag.len);
inter_new(&main_sigma,fft_mag.len);


for (i=1;i<main_simple.len;i++)
{
	main_simple.data[i]=(data.data[i]*data.x[i]);
	main_simple.x[i]=T*kb*log(v0*data.x[i])/1.6e-19;
}

inter_save_a(&main_simple,outputpath,"rod_simple.dat");

double val=0.0;
//out=fopena(outputpath,"./rod_fft_result.dat","w");
for (i=1;i<fft_mag.len;i++)
{
	phase=fft_phase.data[i];
	mag=fft_mag.data[i];

	val=mull*((data.data[0]*sin(phase))/(mag));//-omega_scale.x[i];

	main_exp.data[i]=val;


	main_exp.x[i]=T*kb*log(v0*(2.0*pi)/fft_mag.x[i])/q;
//	fprintf(out,"%le %le\n",main_exp.x[i],val);
}
//fclose(out);

double div=inter_get(&main_exp,0.125);
inter_rescale(&main_exp,1.0, 0.0043/div);

struct istruct outtemp;
inter_copy(&outtemp,&main_exp,TRUE);
div=inter_get(&outtemp,norm_E_pos);
inter_rescale(&outtemp,1.0, 1.0/div);
inter_save_a(&outtemp,outputpath,"rod_fft_result.dat");

inter_free(&outtemp);

inter_mod(&main_exp);

if (dolog==TRUE) inter_log_y(&main_exp);

printf("%d\n",do_fit);

if (do_fit==TRUE)
{

for (i=1;i<fft_mag.len;i++)
{
	//main_exp.data[i]=log10(fabs(main_exp.data[i]));


	main_g.data[i]=1e43;
	main_g.x[i]=main_exp.x[i];
}



window(&main_g,1.0,1.0);
window(&main_g_save,1.0,1.0);
window(&main_fit,1.0,1.0);
window(&main_error,1.0,1.0);
window(&main_exp,1.0,1e-10);
window(&main_sigma,1.0,1.0);
window(&main_simple,1.0,1.0);
inter_save_a(&main_exp,outputpath,"rod_delta_window.dat");
inter_save_a(&main_g,outputpath,"rod_g.dat");
//printf("I am=%ld\n",main_g.len);
//getchar();
char temp[1000];
strcpy(temp,outputpath);
strcat(temp,"g_init.dat");
inter_set_value(&main_g,1e10);
//inter_load_a(&main_g,temp,0,1);

const gsl_multifit_fdfsolver_type *T;
gsl_multifit_fdfsolver *s;
int status;
unsigned iter = 0;
const size_t n = fit_stop-fit_start;
const size_t p = fit_stop-fit_start;

gsl_matrix *covar = gsl_matrix_alloc (p, p);

inter_set_value(&main_sigma,1.0);

struct data d = { n,  main_exp.data, main_sigma.data};

gsl_multifit_function_fdf f;


gsl_vector_view x = gsl_vector_view_array (((main_g.data+fit_start)), p);

//printf("test=%le %le\n",gsl_vector_get(&(x.vector), 1),main_g.data[fit_start+1]);
//getchar();

f.f = &expb_f;
f.df = &expb_df;
f.fdf = &expb_fdf;
f.n = n;
f.p = p;
f.params = &d;


T = gsl_multifit_fdfsolver_lmsder;
s = gsl_multifit_fdfsolver_alloc (T, n, p);
gsl_multifit_fdfsolver_set (s, &f, &x.vector);

do
{
   iter++;
   status = gsl_multifit_fdfsolver_iterate (s);

   printf ("status = %s\n", gsl_strerror (status));

           if (status)
             break;

   status = gsl_multifit_test_delta (s->dx, s->x,1e-10, 1e-10);


inter_save_a(&main_exp,outputpath,"rod_exp.dat");
inter_save_a(&main_fit,outputpath,"rod_fit.dat");
inter_save_a(&main_error,outputpath,"rod_error.dat");
inter_save_a(&main_g,outputpath,"rod_g.dat");

inter_copy(&outtemp,&main_g,TRUE);
div=inter_get(&outtemp,norm_E_pos);
inter_rescale(&outtemp,1.0, 1.0/div);
inter_save_a(&outtemp,outputpath,"rod_g_norm.dat");

inter_free(&outtemp);


printf("~log = %d ~~~~~~~~~~~~~~~~~~~%d~~~~~~~~~~~~~~~~~~~~~~~~~\n",dolog,iter);
//if (getchar()=='l') dolog=TRUE;


}
while (status == GSL_CONTINUE && iter < 10000);

gsl_multifit_covar (s->J, 0.0, covar);

#define FIT(i) gsl_vector_get(s->x, i)
#define ERR(i) sqrt(gsl_matrix_get(covar,i,i))

printf ("status = %s\n", gsl_strerror (status));

gsl_multifit_fdfsolver_free (s);
gsl_matrix_free (covar);



if (donotdump==FALSE)
{
inter_save_a(&main_fit,outputpath,"fit.dat");
inter_save_a(&main_error,outputpath,"error.dat");
}
}


inter_free(&fft_mag);
inter_free(&fft_phase);

inter_free(&main_sigma);
inter_free(&data);
inter_free(&fft);
inter_free(&main_g_save);
inter_free(&main_fit);
inter_free(&main_exp);
inter_free(&main_error);
inter_free(&fft_in);
}

void fft_new(struct istruct* in)
{
inter_new(in,fft_len*2);
}

void fft_to_mag(struct istruct* out,struct istruct* in)
{

int i;
for (i=0;i<in->len/2;i++)
{
	//printf("%le\n",in->data[i*2]);
	//getchar();
	out->data[i]=fft_mag(in,i);

	out->x[i]=in->x[i*2];
}


}

void fft_to_phase(struct istruct* out,struct istruct* in)
{

int i;
for (i=0;i<in->len/2;i++)
{
	out->data[i]=atan(in->data[(i*2)+1]/in->data[(i*2)]);
	out->x[i]=in->x[i*2];
}


}

void fft_to_real(struct istruct* in,struct istruct* out)
{

int i;
for (i=0;i<in->len/2;i++)
{
	out->data[i]=in->data[i*2];
	out->x[i]=in->x[i*2];
}


}

void i_to_fft(struct istruct* out,struct istruct* in)
{
int i=0;
for (i=0;i<in->len;i++)
{
	out->data[i*2]=in->data[i];
	out->x[i*2]=in->x[i];
	out->x[(i*2)+1]=in->x[i];
}

}

void build_dft_energy_mesh(struct istruct* in,double min,double max,double T)
{
int i;
double delta=0.0;
double Estart=min;
double Estop=max;
double Epos=Estart;
delta=(Estop-Estart)/((double)in->len/2.0);
double omega=0.0;

for (i=0;i<in->len/2;i++)
{
	omega=2.0*pi*exp((-Epos*q)/(kb*T))*v0;
	Epos+=delta;
	in->x[i*2]=omega;
	in->x[(i*2)+1]=omega;
	//printf("%le %le\n",Epos,omega);
	//getchar();
}

}


void build_dft_mesh(struct istruct* out,struct istruct* in)
{
int i;
double delta=(in->x[((in->len/2-1)*2)]-in->x[(0*2)])/(out->len/2);
printf("%le\n",delta);
double omega=0.0;

for (i=0;i<out->len/2;i++)
{
	omega=2.0*pi*((double)i/(double)(out->len/2))/delta;
	out->x[i*2]=omega;
	out->x[(i*2)+1]=omega;

}

}

void do_dft(struct istruct* out,struct istruct* in)
{
printf("start dft\n");
int i;
int ii;
double tn=0.0;
double tp=0.0;
double dt=0.0;
double t=0.0;
double omega=0.0;
double a=0.0;
double b=0.0;

for (i=0;i<out->len/2;i++)
{
out->data[i*2]=0.0;
out->data[(i*2)+1]=0.0;
//printf("%le",omega);
//getchar();
//if (i==0) omega=0.0;
	omega=out->x[i*2];
	//printf("%le\n",omega);
	//getchar();
FILE *intout=fopena(global_output,"./out.dat","w");
	for (ii=0;ii<in->len/2;ii++)
	{


		if (ii==0)
		{
			tn=0.0;
		}else
		{
			tn=in->x[(ii*2)-1];
		}

		if (ii==in->len-1)
		{
			tp=in->x[(ii*2)]+(in->x[(ii*2)]-in->x[((ii-1)*2)])/2.0;
		}else
		{
			tp=in->x[(ii*2)+1];
		}

		dt=(tp-tn)/2.0;
		t=in->x[ii*2];


		a=in->data[ii*2];
		b=in->data[(ii*2)+1];
		double inta;
		double intb;
		inta=a*cos(omega*t)*dt;
		intb=a*sin(omega*t)*dt;
		//printf("%le %le\n",omega,t);
		//getchar();
		out->data[i*2]+=inta;
		out->data[(i*2)+1]+=intb;

		fprintf(intout,"%le %le %le\n",omega*t,inta,intb);
		//printf("%le %le %le %le\n",t,omega*t,inta,intb);

		//printf("%le %d %d\n",in->x[i*2],i,ii);
		//getchar();


	}
fclose(intout);
//printf("wait here %le\n",omega);
//getchar();
		//printf("%le %le\n",omega,sqrt(out->data[i*2]*out->data[i*2]+out->data[(i*2)+1]*out->data[(i*2)+1]));
		//getchar();
}
printf("start out\n");
}


