// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>
#include <gsl/gsl_multimin.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "server.h"
#include "fit.h"

static struct fitvars fitconfig;
int uptodate=FALSE;


static double last_iv_error=0.0;
static double last_charge_error=0.0;
static double last_iv_error_light=0.0;
static double last_charge_error_light=0.0;
static double last_rbi_error=0.0;
static double last_tail_error=0.0;
static double last_total_error=0.0;
static double last_tpc_error=0.0;
static double last_tof_error=0.0;
static double last_voc_tau_error=0.0;
static double last_ce_trans_error=0.0;

double get_rbi_error()
{
char orig_file[200];
strcpy(orig_file,"./chris/k.dat");
printf("get iv error\n");
int i;
double error=0.0;
struct istruct exp;
struct istruct iv;
int fit_len=get_file_len(orig_file)-1;
int ivexternal_len=get_file_len("./rbi.dat")-1;
inter_load(&exp,orig_file,fit_len,0,1);
inter_load(&iv,"./rbi.dat",ivexternal_len,0,3);
//inter_save(&iv,"./rodtest.dat");
//getchar();
FILE *ferror;
ferror=fopen("./rbi_error.dat","w");
//inter_print(&exp);
//getchar();
//printf("len = %d %d\n", ivexternal_len,fit_len);
//getchar();
for (i=0;i<iv.len;i++)
{
if (iv.data[i]<0) iv.data[i]=0.0;
iv.data[i]=sqrt(iv.data[i]*iv.data[i]);

}
double max=0.0;
for (i=0;i<exp.len;i++)
{
//printf("%e\n",exp.data[i]);
exp.data[i]=sqrt(exp.data[i]*exp.data[i]);
if (exp.data[i]>max) max=exp.data[i];
}

for (i=0;i<iv.len;i++)
{
if (iv.data[i]>max) max=iv.data[i];
}

for (i=0;i<iv.len;i++)
{
iv.data[i]/=max;
//printf("%e\n",iv.data[i]);
}

for (i=0;i<exp.len;i++)
{
exp.data[i]/=max;
//printf("%e\n",exp.data[i]);
}
//getchar();
double start=exp.x[0];
if (iv.x[0]>start) start=iv.x[0];
double stop=exp.x[exp.len-1];
double points=0.0;
double dx=(stop-start)/1000;
double pos=start;
double etemp=0.0;
do
{
etemp=fitconfig.rbi_fitmul*fabs(inter_get(&iv,pos)-inter_get(&exp,pos));
fprintf(ferror,"%e %e %e %e\n",pos,etemp,inter_get(&iv,pos),inter_get(&exp,pos));
error+=etemp;
points+=1.0;

pos+=dx;
//printf("%e\n",error);
}while(pos<stop);

fclose(ferror);
inter_free(&exp);
inter_free(&iv);









//printf("return = %e\n",error);
//getchar();
return error/points;
}

double get_all_error(struct fitvars *myfit)
{
int i;
double local_error=0.0;
if (myfit->fit_dark_jv==TRUE)
{
last_iv_error=get_iv_error(myfit);
printf("last_iv_error = %e\n",last_iv_error);
}else
{
last_iv_error=0.0;
}

if (myfit->dotimece==TRUE)
{
last_charge_error=get_charge_error(myfit);
}else
if (myfit->fit_dark_ce_ss==TRUE)
{
last_charge_error=get_charge_ss_error(myfit);
printf("last_charge_error = %e\n",last_charge_error);
}else
{
last_charge_error=0.0;
}

if (myfit->dotimecelight==TRUE)
{
last_iv_error_light=get_charge_error_light(myfit);
}else
if (myfit->dolightfit==TRUE)
{
last_iv_error_light=get_iv_error_light(myfit);
printf("last_iv_error_light = %e\n",last_iv_error_light);
}else
{
last_iv_error_light=0.0;
}

if (myfit->fit_light_ce==TRUE)
{
last_charge_error_light=get_charge_light_ss_error(myfit);
printf("last_charge_error_light = %e\n",last_charge_error_light);
}else
{
last_charge_error_light=0.0;
}

last_tof_error=0.0;
for (i=0;i<myfit->tofmax;i++)
{
local_error=get_tof_error(myfit,i);
last_tof_error+=local_error;
printf("last_tof_error = %e\n",local_error);

}

last_ce_trans_error=0.0;
for (i=0;i<myfit->cemax;i++)
{
	local_error=get_ce_trans_error(myfit,i);
	last_ce_trans_error+=local_error;
	printf("last_ce_trans_error = %le\n",local_error);
}

last_tpc_error=0.0;
if (myfit->tpcmax!=0)
{
	for (i=0;i<myfit->tpcmax;i++)
	{
	local_error=get_tpc_error(myfit,i);
	last_tpc_error+=local_error;
	printf("tpc_error %d = %e\n",i,local_error);

	}
	printf("total_tpc_error = %e\n",last_tpc_error);
}

if (myfit->fit_rbi==TRUE)
{
last_rbi_error=get_rbi_error();
printf("last_rbi_error = %e\n",last_rbi_error);
}
else
{
last_rbi_error=0.0;
}

if (myfit->dolightfit2000==TRUE)
{
last_voc_tau_error=0.0;//get_voc_tau_error(myfit);
printf("last_voc_tau_error = %e\n",last_voc_tau_error);
}
else
{
last_voc_tau_error=0.0;
}

last_tail_error=get_tail_error(myfit);

printf("last_general_error = %e\n",last_tail_error);

last_total_error=last_iv_error+last_charge_error+last_rbi_error+last_iv_error_light+last_charge_error_light+last_tail_error+last_tpc_error+last_tof_error+last_voc_tau_error+last_ce_trans_error;

printf("total error = %e\n",last_total_error);

return last_total_error;
}

double my_f (const gsl_vector *v, void *params)
{
double val;
double e0;
int fitcount=0;

for (fitcount=0;fitcount<fitconfig.fitvars;fitcount++)
{
	val = gsl_vector_get(v, fitcount);
	edit_file(fitconfig.fit_file[fitcount],"",fitconfig.fit_pos[fitcount],val);
}

e0=fit_run_sims(&fitconfig,&globalserver);


printf("Done first round\n");
//getchar();
return e0;
}

void my_df (const gsl_vector *v, void *params,
       gsl_vector *df)
{
int i;
double e0=my_f (v,params);
double e1=0.0;
double x0=0.0;
double x1=0.0;
double grad=0.0;
for (i=0;i<v->size;i++)
{
	x0=gsl_vector_get(v, i);
	x1=x0*(1.0+0.01);
	if (x1==x0) x1=x0+1e-6;
	gsl_vector_set((gsl_vector *)v, i, x1);
	e1=my_f (v,params);
	gsl_vector_set((gsl_vector *)v, i, x0);
	e0=my_f (v,params);
	grad=(e1-e0)/(x1-x0);
	gsl_vector_set(df, i, grad);
	printf("J=%d %le %le %le\n",i,grad,x0-x1,e1-e0);

}
printf("Finished grad\n");
}

void my_fdf (const gsl_vector *x, void *params, double *f, gsl_vector *df)
{
  *f = my_f(x, params);
  my_df(x, params, df);

}


void fit_dump_log(int iter,double error,double size,gsl_vector *old_vector,gsl_vector *old_shift,gsl_vector *ss,gsl_vector *x)
{
FILE *out = fopen("./fitlog.dat","a");

printf("---------------------------------------------------------------------\n");

gsl_vector_sub (old_vector, x);
gsl_vector_sub (old_shift, ss);

printf ("%d %le %le %le\n", iter, error, last_total_error,size);
fprintf (out,"%d %le  %le %le %le", iter, error, size,server_get_jobs_per_s(),server_get_odes_per_s());
//fprintf(out,"%e %e %e %e %e %e %e %e %e %e",last_total_error,last_iv_error,last_tpc_error,last_charge_error,last_tail_error,last_iv_error_light,last_charge_error_light,last_rbi_error,last_dark_time_error,last_light_time_error);
//for (i=0;i<item;i++)
//{
//	printf("%e %e %e %e\n",gsl_vector_get (s->x, i),gsl_vector_get(ss, i),gsl_vector_get (old_vector, i),gsl_vector_get (old_shift, i));
//	fprintf(out,"%e ",gsl_vector_get (s->x, i));
//}
//	printf("\n");
fprintf(out,"\n");


printf("---------------------------------------------------------------------\n");

fclose(out);
}

void do_copy()
{
FILE *file;
char path[1000];
char command[1000];

	if ((file=fopen("./rsyncdir.inp","r"))==NULL)
	{
		printf("No rsyncdir.inp file so not copying\n");
		exit(0);
	}else
	{

		fscanf(file,"%s",path);
		fclose(file);
		sprintf(command,"cp ./*.dat ./*.inp %s\n",path);

		printf("Running command %s\n",command);
		system(command);

	}


}

int fit_now()
{
double val=0.0;
double par[5] = {1.0, 2.0, 10.0, 20.0, 30.0};
size_t iter = 0;
int status;
int item=0;
double size;
int fitcount=0;
int loop=0;
double shift;
int cont=TRUE;

fit_read_config(&fitconfig);
shift=fitconfig.simplexmul;
item=fitconfig.fitvars;


gsl_vector *x;
gsl_vector *ss;

gsl_vector *old_vector;
gsl_vector *old_shift;

x = gsl_vector_alloc (fitconfig.fitvars);
ss = gsl_vector_alloc (fitconfig.fitvars);
old_vector = gsl_vector_alloc (fitconfig.fitvars);
old_shift = gsl_vector_alloc (fitconfig.fitvars);

for (fitcount=0;fitcount<fitconfig.fitvars;fitcount++)
{
	val=read_value(fitconfig.fit_file[fitcount],0,fitconfig.fit_pos[fitcount]);
	gsl_vector_set (x, fitcount, val);
	gsl_vector_set (ss, fitcount, fabs(val*shift));
}

if (fitconfig.newton==TRUE)
{
gsl_multimin_fdfminimizer *s_newton = NULL;
	const gsl_multimin_fdfminimizer_type *Tnewton = gsl_multimin_fdfminimizer_vector_bfgs;//gsl_multimin_fminimizer_nmsimplex;//gsl_multimin_fminimizer_nmsimplex;
	gsl_multimin_function_fdf minex_func_newton;

	minex_func_newton.n = item;
	minex_func_newton.f = my_f;
	minex_func_newton.df = my_df;
	minex_func_newton.fdf = my_fdf;

	minex_func_newton.params = par;


	s_newton = gsl_multimin_fdfminimizer_alloc (Tnewton, item);
	//gsl_multimin_fminimizer_set (s_newton, &minex_func_newton, x, ss);
	gsl_multimin_fdfminimizer_set (s_newton, &minex_func_newton, x, 1e-6, 1e-4);

	do
	{
		iter++;
		status = gsl_multimin_fdfminimizer_iterate (s_newton);

		if (status)
		{
			break;
		}
		size = gsl_multimin_test_gradient (s_newton->gradient, 1e-3);
		status = gsl_multimin_test_gradient (s_newton->gradient, 1e-8);

		if (status == GSL_SUCCESS)
		{
			printf ("converged to minimum at\n");
		}

		fit_dump_log(iter,s_newton->f, size,old_vector,old_shift,ss,s_newton->x);

		gsl_vector_memcpy (old_vector, s_newton->x);
		gsl_vector_memcpy (old_shift, ss);

		if (loop>fitconfig.simplexreset) break;
		loop++;

		do_copy();

	}while (status == GSL_CONTINUE);

	printf ("status = %s\n", gsl_strerror (status));


	gsl_multimin_fdfminimizer_free (s_newton);
	cont=FALSE;

}else
{
	const gsl_multimin_fminimizer_type *T =
	gsl_multimin_fminimizer_nmsimplex;//gsl_multimin_fminimizer_nmsimplex;
	gsl_multimin_function minex_func;
	gsl_multimin_fminimizer *s = NULL;


	/* Initialize method and iterate */
	minex_func.n = item;
	minex_func.f = my_f;
	minex_func.params = par;

	s = gsl_multimin_fminimizer_alloc (T, item);
	gsl_multimin_fminimizer_set (s, &minex_func, x, ss);
	double last_error=0.0;
	int length=0;
	int stop=FALSE;
	int dump_count=0;
	do
	{
		status = gsl_multimin_fminimizer_iterate(s);

		if (status) break;

		size = gsl_multimin_fminimizer_size (s);
		status = gsl_multimin_test_size (size, 1e-8);

		if (status == GSL_SUCCESS)
		{
			printf ("converged to minimum at\n");
		}

		fit_dump_log(iter,s->fval, size,old_vector,old_shift,ss,s->x);

		if (loop>fitconfig.simplexreset)
		{
			if (length>fitconfig.fitvars)
			{
				stop=TRUE;
			}else
			{
				loop-=fitconfig.fitvars*2;
			}
		}

		if (last_error==s->fval)
		{
			length++;
		}else
		{
			length=0;
		}

		if (status != GSL_CONTINUE) stop=TRUE;

		if ((dump_count>200)||(stop==TRUE))
		{
			do_copy();
			dump_count=0;
		}

		loop++;
		dump_count++;
		iter++;
		last_error=s->fval;

	}while (stop==FALSE);


	gsl_multimin_fminimizer_free (s);



	cont=TRUE;
}

gsl_vector_free (old_vector);
gsl_vector_free (old_shift);
gsl_vector_free(x);
gsl_vector_free(ss);
return cont;
}
//</clean>
