// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#define _FILE_OFFSET_BITS 64
#define _LARGEFILE_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "i.h"
#include "sim.h"

void inter_add(struct istruct* out,struct istruct* in)
{
int i;
	for (i=0;i<out->len;i++)
	{
		out->data[i]+=in->data[i];
	}

}
void inter_alloc_dd(struct istruct* in,int m,int len)
{
int i=0;
in->m=m;
in->dd = (double **)malloc (in->m*sizeof(double));

for (i=0;i<in->m;i++)
{
	in->dd[i]=(double *)malloc (len*sizeof(double));
}

in->data=in->dd[0];
}

void inter_to_log_mesh(struct istruct* out,struct istruct* in)
{
double a=log10(in->x[0]);
double b=log10(in->x[in->len-1]);
double step=(b-a)/((double)out->len);
int i;
double pos=a;
for (i=0;i<out->len;i++)
{
out->x[i]=pow(10.0,pos);
out->data[i]=inter_get(in,pow(10.0,pos));
pos+=step;

}

}

void inter_to_new_mesh(struct istruct* in,struct istruct* out)
{
int i;
int ii;
double pos=0.0;
double delta=(in->x[in->len-1]-in->x[0])/(double)out->len;
pos+=delta/2.0;
for (i=0;i<out->len;i++)
{
	ii=search(in->x,in->len,pos);

	double x0=in->x[ii];
	double x1=in->x[ii+1];

	double y0=in->data[ii];
	double y1=in->data[ii+1];

	out->x[i]=pos;
	out->data[i]=y0+((y1-y0)/(x1-x0))*(pos-x0);

	//printf("%d %d %le %le %le %le \n",i,ii,out->x[i],out->data[i],delta,in->x[i]);
	pos+=delta;
}

return;
}


void inter_realloc_dd(struct istruct* in,int len)
{
int m;
for (m=0;m<in->m;m++)
{
	in->dd[m]=(double *)realloc (in->dd[m],len*sizeof(double));
}

in->data=in->dd[0];
}

void inter_new(struct istruct* in,int len)
{
int i;
in->m=1;
strcpy(in->name,"new");

in->len=len;

in->x = (double *)malloc (in->len*sizeof(double));

inter_alloc_dd(in,in->m,in->len);

for  (i=0;i<in->len;i++)
{
	in->data[i]=0.0;
}

}

double inter_sum_mod(struct istruct* in)
{
int i;
double sum=0.0;

for (i=0;i<in->len;i++)
{
	sum+=fabs(in->data[i]);
}
return sum;
}

double inter_sum(struct istruct* in)
{
int i;
double sum=0.0;

for (i=0;i<in->len;i++)
{
	sum+=in->data[i];
}
return sum;
}

double inter_get_diff(struct istruct* one,struct istruct* two,char *filename,double start,double stop,double error_mul)
{
FILE *ferror;
ferror=fopen(filename,"w");
double error=0.0;
if (one->x[0]>start) start=one->x[0];

double points=0.0;
double dx=(stop-start)/1000;
double pos=start;
double etemp=0.0;
do
{

	if (pos>=start)
	{
		etemp=error_mul*fabs(inter_get(one,pos)-inter_get(two,pos));
		fprintf(ferror,"%e %e %e %e\n",pos,etemp,inter_get(one,pos),inter_get(two,pos));
		error+=etemp;
		points+=1.0;
	}

	pos+=dx;
}while(pos<stop);

fclose(ferror);

//inter_save(one,"./a.dat");
//inter_save(two,"./b.dat");

return error/points;
}

double inter_intergrate(struct istruct* in)
{
int i;
double tn=0.0;
double tp=0.0;
//double t=0.0;
double dt=0.0;
//double Eomega=0.0;
double sum=0.0;
double n;

	for (i=0;i<in->len;i++)
	{

		if (i==0)
		{
			tn=in->x[i];
		}else
		{
			tn=in->x[i-1];
		}

		if (i==in->len-1)
		{
			tp=in->x[i];
		}else
		{
			tp=in->x[i+1];
		}

		n=in->data[i];
		dt=fabs((tp-tn)/2.0);

		sum+=n*dt;


	}

return sum;
}

double inter_norm_to_one_range(struct istruct* in,double start,double stop)
{
int i;
double max=0.0;

for (i=0;i<in->len;i++)
{
	if (in->x[i]>start)
	{
		max=in->data[i];
		break;
	}
}

for (i=0;i<in->len;i++)
{
	if ((in->x[i]>start)&&(in->x[i]<stop))
	{
		if (in->data[i]>max) max=in->data[i];
	}
}

for (i=0;i<in->len;i++)
{
in->data[i]/=max;
}

return max;
}

double inter_norm_to_one(struct istruct* in)
{
int i;
double max=in->data[0];
//if (in->len>0) max=in->data[0];
for (i=0;i<in->len;i++)
{
if (in->data[i]>max) max=in->data[i];
}

for (i=0;i<in->len;i++)
{
in->data[i]/=max;
}

return max;
}

void inter_log_y_m(struct istruct* in)
{
int i;
double mull=1.0;

for (i=0;i<in->len;i++)
{
	mull=1.0;
	if (in->data[i]<0.0) mull=-1.0;
	in->data[i]=log10(sqrt(in->data[i]*in->data[i]))*mull;
}
}

void inter_log_y(struct istruct* in)
{
int i;
for (i=0;i<in->len;i++)
{
	in->data[i]=log10(sqrt(in->data[i]*in->data[i]));
}
}

void inter_log_x(struct istruct* in)
{
int i;
for (i=0;i<in->len;i++)
{
	if (in->x[i]>0.0) in->x[i]=log10(sqrt(in->x[i]*in->x[i]));
}
}

void inter_smooth(struct istruct* in,int points)
{
int i=0;
int ii=0;
int pos=0;
double tot_point=0.0;
double tot=0;
	for (i=0;i<in->len;i++)
	{
		for (ii=0;ii<points;ii++)
		{
			pos=i+ii;
			if (pos<in->len)
			{
				tot+=in->data[pos];
				tot_point+=1.0;
			}
		}

		in->data[i]=(tot/(double)tot_point);
		tot=0.0;
		tot_point=0.0;
	}
}

void inter_purge_zero(struct istruct* in)
{
int i;
int write=0;
int read=0;
for (i=0;i<in->len;i++)
{
	in->data[write]=in->data[read];
	in->x[write]=in->x[read];
	if (in->data[read]==0.0)
	{
		write--;
	}
	read++;
	write++;
}
in->len=write;

in->x = (double *)realloc (in->x,in->len*sizeof(double));
inter_realloc_dd(in,in->len);

}

double inter_get_min(struct istruct* in)
{
int i=0;
double min=in->data[i];
for (i=0;i<in->len;i++)
{
	if (in->data[i]<min) min=in->data[i];


}
return min;
}

void inter_chop(struct istruct* in,double min, double max)
{
int i;
int write=0;
int read=0;
for (i=0;i<in->len;i++)
{
	in->data[write]=in->data[read];
	in->x[write]=in->x[read];
	write++;
	if (in->x[read]<min)
	{
		write--;
	}

	if (in->x[read]>max) break;
	read++;




}
in->len=write;

in->x = (double *)realloc (in->x,in->len*sizeof(double));
inter_realloc_dd(in,in->len);
}



/**Rescale the scale and the data
@param in The structure holding the data
@param xmul multiply x axis by this
@param ymul multiply y axis by this
*/

void inter_rescale(struct istruct* in,double xmul, double ymul)
{
int i;
for  (i=0;i<in->len;i++)
{
	in->x[i]*=xmul;
	in->data[i]*=ymul;
}

}

/**Make all the data positive
@param in the structure holding the data
*/
void inter_mod(struct istruct* in)
{
int i;
for  (i=0;i<in->len;i++)
{
if (in->data[i]<0.0) in->data[i]*=-1.0;
}

}

void inter_sub(struct istruct* in,double value)
{
int i;
for  (i=0;i<in->len;i++)
{
in->data[i]-=value;
}

}

void inter_norm(struct istruct* in,double mul)
{
int i;
double tot=0.0;
double dx=0.0;
for  (i=0;i<in->len;i++)
{
	if (i==0)
	{
		dx=in->x[1]-in->x[0];
	}else
	if (i==in->len-1)
	{
		dx=in->x[i]-in->x[in->len-2];
	}else
	{
		dx=(in->x[i+1]-in->x[i])/2.0+(in->x[i]-in->x[i-1])/2.0;
	}

	tot+=dx*in->data[i];
}

for  (i=0;i<in->len;i++)
{

	in->data[i]/=tot;
	in->data[i]*=mul;

}

}

void inter_append(struct istruct* in,double x,double y)
{
in->x[in->len]=x;
in->data[in->len]=y;
in->len++;

if ((in->max_len-in->len)<10)
{
	in->max_len+=100;
	in->x = (double *)realloc (in->x,in->max_len*sizeof(double));
	inter_realloc_dd(in,in->max_len);
}

}




void inter_init_m(struct istruct* in,int m)
{
in->len=0;
in->max_len=100;
in->m=m;
in->x = (double *)malloc (in->max_len*sizeof(double));

inter_alloc_dd(in,m,in->max_len);
}


void inter_init(struct istruct* in)
{
in->m=1;
inter_init_m(in,in->m);
}

void inter_load_a(struct istruct* in,char *name,int col,int ncol)
{
in->m=1;
strcpy(in->name,name);
int len=get_file_len(name);
FILE *file;
file=fopen(name,"r");
if (file == NULL)
{
	printf("inter_load_a can not open file %s\n",name);
	exit(0);
}

in->len=len;
int i=0;
double out=0.0;
in->x = (double *)malloc (in->len*sizeof(double));

inter_alloc_dd(in,in->m,in->len);
int ii=0;
for  (i=0;i<in->len;i++)
{
	fscanf(file,"%lf",&(in->x[i]));
	for (ii=0;ii<ncol;ii++)
	{
		fscanf(file,"%lf",&(out));
		if (ii==col) in->data[i]=out;
	}

}

fclose(file);
}

void inter_copy(struct istruct* in,struct istruct* orig,int alloc)
{
int i;
in->m=orig->m;
in->len=orig->len;

if (alloc==TRUE)
{
in->x = (double *)malloc (orig->len*sizeof(double));
inter_alloc_dd(in,in->m,orig->len);
}

for  (i=0;i<orig->len;i++)
{
	in->x[i]=orig->x[i];
	in->data[i]=orig->data[i];
}


}

void inter_swap(struct istruct* in)
{
int i;
double *xtemp=malloc(sizeof(double)*in->len);
double *dtemp=malloc(sizeof(double)*in->len);


for  (i=0;i<in->len;i++)
{
	dtemp[i]=in->data[i];
	xtemp[i]=in->x[i];
}

for  (i=0;i<in->len;i++)
{
in->data[i]=dtemp[in->len-1-i];
in->x[i]=xtemp[in->len-1-i];
}

free(xtemp);
free(dtemp);
}

/**Load data from a file
@param in the structure holding the data
@param name The file name.
*/
void inter_load(struct istruct* in,char *name,int len,int col,int ncol)
{
in->m=1;
FILE *file;
file=fopen(name,"r");
in->len=len;
strcpy(in->name,name);
int i=0;
double out=0.0;
in->x = (double *)malloc (in->len*sizeof(double));
inter_alloc_dd(in,in->m,in->len);
int ii=0;
for  (i=0;i<in->len;i++)
{
	fscanf(file,"%lf",&(in->x[i]));
	for (ii=0;ii<ncol;ii++)
	{
		fscanf(file,"%lf",&(out));
		if (ii==col) in->data[i]=out;
	}

}

fclose(file);
}

void inter_set_value(struct istruct* in,double value)
{
int i=0;
int m=0;
for  (i=0;i<in->len;i++)
{
	in->dd[m][i]=value;
}

}

void inter_dump(struct istruct* in)
{
int i=0;
int m=0;
for  (i=0;i<in->len;i++)
{
	printf("%e",in->x[i]);
	for (m=0;m<in->m;m++)
	{
		printf(" %e",in->dd[m][i]);
	}
printf("\n");
}

}

void inter_save_a(struct istruct* in,char *path,char *name)
{
char wholename[200];
sprintf(wholename,"%s/%s",path,name);
inter_save(in,wholename);
}

void inter_save_seg(struct istruct* in,char *path,char *name,int seg)
{
FILE *file;
int i=0;
int m=0;
//printf("%d %d\n",in->m,in->len);
//getchar();
int count_max=in->len/seg;
int count=0;
char temp[1000];
int file_count=0;
for  (i=0;i<in->len;i++)
{
if (count==0)
{

sprintf(temp,"%s/%s%d.dat",path,name,file_count);
//printf("%s\n",temp);
file=fopen(temp,"w");
file_count++;
}
	fprintf(file,"%e",in->x[i]);
	for (m=0;m<in->m;m++)
	{
		fprintf(file," %e",in->dd[m][i]);
	}
count++;
fprintf(file,"\n");
if (count==count_max)
{
fclose(file);
count=0;
}

}
if (count!=0) fclose(file);

}



void inter_save(struct istruct* in,char *name)
{
FILE *file;
file=fopen(name,"w");
int i=0;
int m=0;
//printf("%d %d\n",in->m,in->len);
//getchar();
for  (i=0;i<in->len;i++)
{
	fprintf(file,"%e",in->x[i]);
	for (m=0;m<in->m;m++)
	{
		fprintf(file," %e",in->dd[m][i]);
	}
fprintf(file,"\n");
}

fclose(file);
}


/**Get interpolated data from a data set
@param in The structure holding the data
@param x the position of the data.
@return the interpolated data value
*/
double inter_get(struct istruct* in,double x)
{
double x0;
double x1;
double y0;
double y1;

double ret;
int i=0;

//if (x>in->x[in->len-1]) return 0.0;
if (x<in->x[0])
{

return 0.0;
}


/*for (i=0;i<in->len;i++)
{
	if (in->x[i]>x)
	{
		break;
	}
}

i--;
if (i<0) i=0;

if (i>=(in->len-1)) i=in->len-2;
*/

if (x>=in->x[in->len-1])
{
//inter_dump(in);
//printf("%s\n",in->name);
//getchar();
	i=in->len-1;
	x0=in->x[i-1];
	x1=in->x[i];

	y0=in->data[i-1];
	y1=in->data[i];

}else
{
	i=search(in->x,in->len,x);
	x0=in->x[i];
	x1=in->x[i+1];

	y0=in->data[i];
	y1=in->data[i+1];
}
ret=y0+((y1-y0)/(x1-x0))*(x-x0);
return ret;
}

double inter_get_hard(struct istruct* in,double x)
{
//double x0;
//double x1;
//double y0;
//double y1;

//double ret;
//int i=0;

if (x>in->x[in->len-1])
{

return 0.0;
}

return inter_get(in,x);
}

/**Free the structure holding the data
@param in The structure holding the data
*/
void inter_free(struct istruct* in)
{
int m=0;
in->len=0;
in->max_len=0;
free(in->x);
for (m=0;m<in->m;m++)
{
	free(in->dd[m]);
}
free(in->dd);
in->m=0;
}
