// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>

void my_guess(struct device *in)
{
int i;
#ifndef _fast_
printf("stats = %d",in->bdos);
#endif
//getchar();
int stats=in->bdos;
double Ef=0.0;
double charge_left=fabs(in->mat.l[0].space);
double charge_right=fabs(in->mat.l[in->ymatlayers[in->ymeshpoints-1]].space);

double top_l=0.0;
double top_r=0.0;
if (in->interfaceleft==interface_schottky)
{
	top_l=-in->phibleft;
}else
{
	if (in->lcharge==1)
	{
	top_l=get_top_from_p(charge_left,in->Te[0],in->mh[0],in->Nv[0],stats);
	}else
	{
	top_l=-(in->Eg[0]+get_top_from_n(charge_left,in->Th[0],in->me[0],in->Nc[0],stats));
	}
}

if (in->interfaceright==interface_schottky)
{
	top_r=-in->phibright;
}else
{
	if (in->rcharge==1)
	{
		top_r=get_top_from_n(charge_right,in->Te[in->ymeshpoints-1],in->me[in->ymeshpoints-1],in->Nc[in->ymeshpoints-1],stats);
	}else
	{
		top_r=-(in->Eg[0]+get_top_from_p(charge_right,in->Th[in->ymeshpoints-1],in->mh[in->ymeshpoints-1],in->Nv[in->ymeshpoints-1],stats));
	}
}
#ifndef _fast_
printf("check1= %e %e\n",get_p_den(top_l,in->Te[0],in->mh[0],in->bdos),charge_left);
printf("check2= %e %e\n",get_n_den(top_r,in->Te[in->ymeshpoints-1],in->me[in->ymeshpoints-1],in->bdos),charge_right);
#endif
//printf("%e %e %e %e %e",get_n_den(top_l,in->Te[0],in->mh[0],in->bdos,dos_all),get_n_den(top_l,in->Te[0],in->mh[0],in->bdos,dos_mob),get_n_den(top_l,in->Te[0],in->mh[0],in->bdos,1),get_n_trap_den(top_l,in->Te[0],in->mh[0],in->bdos,0),get_n_den(top_l,in->Te[0],in->mh[0],in->bdos,dos_mob)+get_n_trap_den(top_l,in->Te[0],in->mh[0],in->bdos,0) );
//getchar();

//getchar();
double delta_phi=top_l+top_r+in->Eg[0];
#ifndef _fast_
printf("my check %lf\n",delta_phi);
#endif

Ef=-(top_l+in->Xi[0]+in->phi[0]+in->Eg[0]);
//printf("Ef=%e",Ef);
//getchar();

#ifndef _fast_
printf ("Lp = %e\n", get_p_den((-in->Xi[0]-in->phi[0]-in->Eg[0])-Ef,in->Th[0],in->mh[0],in->bdos));
printf ("Ln = %e\n", get_n_den(Ef-(-in->Xi[0]-in->phi[0]),in->Te[0],in->me[0],in->bdos));

printf ("Rp = %e\n", get_p_den((-in->Xi[in->ymeshpoints-1]-delta_phi-in->Eg[in->ymeshpoints-1])-Ef,in->Th[in->ymeshpoints-1],in->mh[in->ymeshpoints-1],in->bdos));
printf ("Rn = %e\n", get_n_den(Ef-(-in->Xi[in->ymeshpoints-1]-delta_phi),in->Te[in->ymeshpoints-1],in->me[in->ymeshpoints-1],in->bdos));
#endif

//getchar();
//=get_top_from_p(charge_left,in->Te[0],in->mh[0],in->Nv[0],stats,dos_all)-in->Xi[i]-in->phi[i];
for (i=0;i<in->ymeshpoints;i++)
{

	in->phi[i]=0;
	in->Ec[i]=-in->phi[i]-in->Xi[i];
	in->Ev[i]=-in->phi[i]-in->Xi[i]-in->Eg[i];

	in->Fi[i]=Ef;

	in->Fn[i]=Ef;
	in->Fp[i]=Ef;


	in->n[i]=in->Nc[i]*exp(((in->Fi[i]-in->Ec[i])*q)/(kb*in->Te[i]));
	in->p[i]=in->Nv[i]*exp(((in->Ev[i]-in->Fi[i])*q)/(kb*in->Th[i]));
//printf("%e %e\n",in->Fi[i],in->Ec[i]);
	in->mun[i]=get_n_mu();
	in->mup[i]=get_p_mu();


}

in->Vl=0.0;
in->Vr=delta_phi;
printf("%e %e %e\n",in->Vl,in->Vr,delta_phi);
//getchar();
}

void get_initial(struct device *in)
{
my_guess(in);
return;
}

