// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>


void sim_jv(struct device *in)
{
double V=in->Vstart;
double Vstop=in->Vstop;
double Vstep=in->Vstep;
FILE *iv;
int ittr=0;
double J;
double Pden;
int first=TRUE;
double Vlast;
double Jlast;
double Pdenlast;
double Vexternal;
struct istruct ivexternal;
inter_init(&ivexternal);

struct istruct charge;
inter_init(&charge);

struct istruct klist;
inter_init(&klist);

#ifndef mindump
iv=fopena(in->outputpath,"./jv.dat","w");
fclose(iv);

iv=fopena(in->outputpath,"./avgmu.dat","w");
fclose(iv);

iv=fopena(in->outputpath,"./avgmu.dat","w");
fclose(iv);

iv=fopena(in->outputpath,"./freetrapped.dat","w");
fclose(iv);

iv=fopen("./k3d.dat","w");
fclose(iv);
#endif


struct map frame;

frame_init(&frame,in->ymeshpoints,1000,in->ymesh[0],in->ymesh[in->ymeshpoints-1],-6.0,-4.0);
double k_voc=0.0;
double n_voc=0.0;


	do
	{

		in->Vapplied=V;



		solve_all(in);

		J=get_equiv_J(in);
		Vexternal=get_equiv_V(in);

		#ifndef mindump
		iv=fopen("./k3d.dat","a");
		fprintf(iv, "%le %le %le\n",Vexternal, in->Psun/1000.0,get_avg_k(in));
		fclose(iv);
		#endif


		frame_reset(&frame);
		frame_build(&frame,in);
		frame_dump(&frame,in);
		//getchar();

		if (ittr>0)
		{

			inter_append(&ivexternal,get_equiv_V(in),get_equiv_J(in));

		}

		ittr++;
		//iv=fopena(in->outputpath,"./freetrapped.dat","a");




		//fprintf(iv,"%e %e %e %e %e\n",V,get_total_n_trapped_charge(in),get_total_n_charge(in),get_total_p_trapped_charge(in),get_total_p_charge(in));
		//fclose(iv);


		inter_append(&charge,get_equiv_V(in),get_extracted_np(in));


			Pden=fabs(J*Vexternal);




			plot_now(in,"plot");


			if (first==FALSE)
			{

				//check if we have crossed 0V
				if ((Vlast<=0)&&(V>=0.0))
				{
					in->Jsc=Jlast+(J-Jlast)*(0-Vlast)/(V-Vlast);
					printf("Jsc = %e\n",in->Jsc);
				}

				if ((Jlast<=0)&&(J>=0.0))
				{
					in->Voc=Vlast+(Vexternal-Vlast)*(0-Jlast)/(J-Jlast);
					printf("Voc = %e\n",in->Voc);
					FILE *voc_file;
					voc_file=fopena(in->outputpath,"./voc.dat","w");
					fprintf(voc_file,"%e\n",in->Voc);
					fclose(voc_file);
					k_voc=get_avg_recom(in)/pow(get_extracted_np(in),2.0);
					n_voc=get_extracted_np(in);

				}

				if ((Pden>Pdenlast)&&(Vexternal>0.0)&&(J<0.0))
				{
					in->Pmax=Pden;
					in->Pmax_voltage=Vexternal;
				}

				if (Vexternal>Vstop) break;

			}

			#ifndef _fast_
			printf("V=%lf %lf current = %e mA (%e A/m^2) %le\n",V,Vexternal,get_I(in)/1e-3,J,get_J(in));
			#endif

			if (in->stop_start==TRUE) getchar();


			Jlast=J;
			Vlast=Vexternal;
			Pdenlast=Pden;
			first=FALSE;

			V+=Vstep;
			//dialog_set_progress ((in->Vstart+V)/(in->Vstop-in->Vstart));
			if ((Vstep>=0)&&(V>Vstop))
			{
				in->stop=TRUE;
				printf("--------------1\n");
			}

			if ((Vstep<0)&&(V<Vstop))
			{
				in->stop=TRUE;
				printf("--------------2\n");
			}
//printf("%e\n",Vstop);



		if (in->stop==TRUE)
		{
			break;
		}
		inter_append(&klist,get_extracted_np(in),get_avg_recom(in)/(pow(get_extracted_np(in),2.0)));

	}while(1);

in->FF=fabs(in->Pmax/(in->Jsc*in->Voc));

char out[200];
sprintf(out,"Voc= %lf (V)\n",in->Voc);
printdebug( out);
sprintf(out,"Jsc= %lf (A/m^2)\n",in->Jsc);
printdebug( out);
sprintf(out,"Pmax= %lf (W/m^2)\n",in->Pmax);
printdebug( out);
sprintf(out,"Voltage to get Pmax= %lf (V)\n",in->Pmax_voltage);
printdebug( out);
sprintf(out,"FF= %lf\n",in->FF*100.0);
printdebug( out);

sprintf(out,"Efficiency= %lf percent\n",mod(in->Pmax/in->Psun)*100.0);
printdebug( out);

#ifndef mindump
iv=fopen("./ff.dat","w");
fprintf(iv,"%le\n",in->FF*100.0);
fclose(iv);

FILE *outfile=fopen("./eff.dat","w");
fprintf(outfile,"%le",mod(in->Pmax/(in->Psun))*100.0);
fclose(outfile);

inter_save_a(&klist,in->outputpath,"k.dat");
inter_free(&klist);
#endif

FILE *kvocfile=fopena(in->outputpath,"./kvoc.dat","w");
fprintf(kvocfile,"%le %le\n",n_voc,k_voc);
fclose(kvocfile);

kvocfile=fopena(in->outputpath,"./voc_tau.dat","w");
fprintf(kvocfile,"%le %le\n",in->Voc,1.0/(k_voc*n_voc));
fclose(kvocfile);

kvocfile=fopena(in->outputpath,"./n_tau.dat","w");
fprintf(kvocfile,"%le %le\n",n_voc,1.0/(k_voc*n_voc));
fclose(kvocfile);

inter_save_a(&charge,in->outputpath,"charge.dat");
inter_free(&charge);

inter_save_a(&ivexternal,in->outputpath,"ivexternal.dat");
inter_free(&ivexternal);

frame_free(&frame);
}

