#define BUILD_PATH "/home/roderick/solarsim/"
#ifndef sim_h
#define sim_h
#include "const.h"
#include "server.h"
#include "i.h"
#include "newton.h"

#define mat_Xi		0
#define mat_Eg		1
#define mat_mue		2
#define mat_muh		3
#define mat_me		4
#define mat_mh		5
#define mat_epsilonr	6
#define mat_Dex		7
#define mat_kf		8
#define mat_a		9
#define mat_kl		10
#define mat_Xiphoton	11
#define mat_Egphoton	12
#define mat_taue0	13
#define mat_taue1	14
#define mat_tauh0	15
#define mat_tauh1	16
#define mat_A		17
#define mat_deltaE	18
#define mat_capeff      19
#define mat_br		20
#define mat_dostype	21
#define mat_rbi		22

//#define dumpconverge

#define dos_all		-1
#define dos_mob		0
#define dos_trap	2
#define dos_mb		0
#define dos_gau		1
#define dos_fd		2
#define dos_exp_fd 	3
#define dos_exp 	4
#define dos_read 	5

#define sim_mode_jv	0

//<clean>
#define sim_mode_tpv	1
#define sim_mode_ce	2
#define sim_mode_tpc	4
#define sim_mode_li	5
#define sim_mode_findvoc 7
#define sim_mode_jvone 	8
#define sim_mode_tof 	11
#define sim_mode_fft 	12
#define sim_mode_celiv 	14
#define sim_mode_fish	17
#define sim_mode_findlight_for_voc 18
//</clean>

#define tpv_mode_light 0
#define tpv_mode_voc 1

#define srh_1	1
#define srh_2	2
#define srh_3	3
#define srh_4	4

#define interface_schottky	 1
#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif


#ifdef _quick_
    #define    dprintf(f,...)
#else
    #define    dprintf(f,...)    printf(f,##__VA_ARGS__)
#endif




struct mesh
{
double number;
double len;
double den;
};

struct param
{
	int used;
	int len;
	double *c;
};

struct layer
{
double height;
double Nad;
double space;
char name[100];
struct param mat[100];
};

struct layers
{
int number;
struct layer* l;
};



struct device
{
	double Jnleft;
	double Jnright;
	double Jpleft;
	double Jpright;
	int fast;
	double *phi;
	double *Nad;
	double *G;
	double *Gex;
	double *n;
	double *p;
	double *dn;
	double *dp;
	double *Eg;
	double *Xi;
	double *Ev;
	double *Ec;

	double *mun;
	double *mup;

	double *epsilonr;

	double *Fn;
	double *Fp;
	double *Nc;
	double *Nv;
	double *me;
	double *mh;
	double *Tl;
	double *Te;
	double *Th;
	double *ymesh;
	double *R;
	double *Fi;
	int *ymatlayers;
	double *Jn;
	double *Jp;
	double Vapplied;
	int ymeshpoints;
	double Vl;
	double Vr;
	double *x;
	double *t;
	double *xp;
	double *tp;
	double *b;
	double *kf;
	double *kd;
	double *kr;
	double *a;
	double *Rbi;
	double *Rn;
	double *Rp;
	double *kl;
	double *ke;
	double *kh;
	double *Hl;
	double *He;
	double *Hh;
	double *Hasorb;
	double Psun;
	int dd_conv;
	int excite_conv;
	int thermal_conv;

	double deltaFln;
	double deltaFlp;
	double deltaFrn;
	double deltaFrp;

//back tracking
	double *back_x;
	double *back_xp;
	double *back_phi;
	double *back_xt;
	double *back_xpt;
	double *back_b;
	double *Rbi_k;
//back tracking end

	double *ex;
	double *Dex;
	double *Hex;

	double *nfinit;
	double *pfinit;
	double *ntinit;
	double *ptinit;
	double **ntbinit;
	double **ptbinit;

	double xlen;
	double ylen;
	double zlen;

	double kTq;

	struct layers mat;
	struct layer lcontact;
	struct layer rcontact;
	
	int N;
	int M;
	int *Ti;	//row
	int *Tj;	//col
	double *Tx;	//data
	int plot;
	int enable_back;

//plotting
	FILE *gnuplot;
	FILE *gnuplot_time;
	FILE *converge;
	FILE *tconverge;
	int max_electrical_itt;
	double electrical_clamp;
	double min_cur_error;
	double Vstart;
	double Vstop;
	double Vstep;
	int Rbi_on;
	int Gp_on;
	int Gn_on;
	double Pmax_voltage;

//Device characterisation
	double Voc;
	double Jsc;
	double FF;
	double Pmax;
	double Tll;
	double Tlr;
	int Tliso;
	double dt;
	int srh_sim;
	int go_time;
	double *nlast;
	double *plast;
	int tsteps;
	double time;
	int pulsestart;
	int pulselen;
	int reprate;
	int ntrapnewton;
	int ptrapnewton;

	int stop;
	double Rshort;
	double Rshort_celiv;
	double timeshift;
	struct mesh *meshdata;
	int ymeshlayers;
	int onlypos;
	int odes;
	double last_error;
	double posclamp;
	int srh_bands;
	double *wn;
	double *wp;

//n traps
	double *nt_all;
	double **nt;
	double **ntlast;
	double **dnt;
	double **srh_n_r1;
	double **srh_n_r2;
	double **srh_n_r3;
	double **srh_n_r4;
	double **dsrh_n_r1;
	double **dsrh_n_r2;
	double **dsrh_n_r3;
	double **dsrh_n_r4;
	double **Fnt;
	double *Nct;
	double **xt;
	double *tt;

//p traps
	double *pt_all;
	double **pt;
	double **ptlast;
	double **dpt;
	double **srh_p_r1;
	double **srh_p_r2;
	double **srh_p_r3;
	double **srh_p_r4;
	double **dsrh_p_r1;
	double **dsrh_p_r2;
	double **dsrh_p_r3;
	double **dsrh_p_r4;
	double **Fpt;
	double *Nvt;
	double **xpt;
	double *tpt;

	double A;
	double Vol;
	double *deltaE;
	double *capeff;

	double Rshunt;
	double Rcontact;
//ND filter
	double ND;
	int Dphoton;
	double Dphotoneff;
	int bdos;
	char spectra[100];

	int *dostype;

	int interfaceleft;
	int interfaceright;
	double phibleft;
	double phibright;
	double vl;
	double vr;
	int stop_start;
	int laser;
	double wavelength;
	double externalv;
	double Vapplied_last;
	double Ilast;
	int simwire;
	int plottime;
	int timedumpcount;
	char outputpath[200];
	int simmode;
	double area;
	double tpcstop;
	double tpclaseroff;
	double tpcdtstart;
	double tpcdtmull;
	double tpclaserpower;

	double *nrelax;
	double *ntrap_to_p;
	double *prelax;
	double *ptrap_to_n;

	double tpvdtstart;
	double tpvlaseroff;
	double tpvlaserpower;
	double tpvdtmull;
	double tpvstop;
	double tpvvoc;
	int tpvpulsestart;
	double cedtmull;
	double cetimestop;
	double cevoltage;
	double ceshortvoltage;
	int tpv_mode;
	int cefast;
	int lcharge;
	int rcharge;

	double tofstop;
	double toflaseroff;
	double tofdtstart;
	double tofdtmull;
	double toflaserpower;

	double celiv_dt;
	double celiv_t0;
	double celiv_t1;
	double celiv_t2;
	double celiv_Vexternal;
	double celiv_Vmax;
	int celiv_repeat;
	double celiv_time_j0;
	double celiv_time_j1;
	int dumpitdos;
	double cedt;
	int cepresteps;

	double fish_1start;
	double fish_1stop;
	double fish_1den;
	double fish_2start;
	double fish_2stop;
	double fish_2den;
	double fish_externalv;
	double fish_dt;
	double fish_dt_mull;
	double fish_stop;
	double C;

	int newton_cur;
	void (*newton_aux)(struct device* ,double ,double* ,double* ,double* ,double* ,double* ,double* ,double* ,double*);
};


#include "dos.h"

struct map
{
double **data;
int xpoints;
int ypoints;
double xstart;
double ystart;
double xstop;
double ystop;
double xdelta;
double ydelta;
double *x;
double *y;
int count;
};



void get_avg_np_pos(struct device *in,double *nx,double *px);

double get_background_charge(struct device *in);
void save_state(struct device *in,double to);
int load_state(struct device *in,double voltage);
void pt_extract(char *outputpath,char *infile,double T,int pt_fit,double start,double stop,double mull,double dosstart,double dosstop,int maxittr,double chop,double chop1,double norm_point,double norm_E_pos,double sigmain,double v0,double norm_E_mull);
void sim_celiv(struct device *in);
void build_dft_mesh(struct istruct* out,struct istruct* in);
void build_dft_energy_mesh(struct istruct* in,double min,double max,double T);

void fft_extract();
void do_dft(struct istruct* out,struct istruct* in);
void fft_new(struct istruct* in);
void fft_to_mag(struct istruct* in,struct istruct* out);
void fft_to_phase(struct istruct* in,struct istruct* out);
void fft_to_real(struct istruct* in,struct istruct* out);
void i_to_fft(struct istruct* out,struct istruct* in);
void do_fft(struct istruct* in, int isign);


void ramp(struct device *in,double from,double to,double steps);
void dump_device_map(struct device *in);
void sim_jv_one_point(struct device *in);
void set_light_for_voc(struct device *in,double Voc);
void reset_npinit(struct device *in);
void light_update(struct device *in,double value);
void join_tpv_to_ce(struct device *in);
double get_n_trapped_charge(struct device *in);
double get_p_trapped_charge(struct device *in);
double get_avg_recom(struct device *in);
double get_avg_recom_n(struct device *in);
double get_avg_recom_p(struct device *in);
double get_tpc_t0(char * file,double start0,double stop0,double start1,double stop1);
double dos_get_band_energy_n(int band);
double dos_get_band_energy_p(int band);
void frame_build(struct map *mapin,struct device *in);
void frame_reset(struct map *in);
void frame_free(struct map *in);
void frame_init(struct map *in,int xpoints,int ypoints,double xstart,double xstop,double ystart,double ystop);
void frame_dump(struct map *in,struct device *dev);
void frame_add_data(struct map *in,double x,double y,double data);
void sim_li(struct device *in);
double get_J_recom_n(struct device *in);
double get_J_recom_p(struct device *in);
double get_avg_k(struct device *in);

double dos_srh_get_fermi_p(double n, double p,int band);
double dos_srh_get_fermi_n(double n, double p,int band);
void time_init(struct device *in);
void update_arrays(struct device *in);
void tau_test(struct device *in);
void gen_dos_fd_gaus_n();
void gen_dos_fd_gaus_p();


double sim_ce(struct device *in);
void process_ce_data(char *input,char *output);
double newton_charge_extraction(struct device *in);
double newton_celiv(struct device *in,double Vtot,int usecap);
double get_I_ce(struct device *in);
double get_equiv_I(struct device *in);
void sim_tpc(struct device *in);
void sim_tof(struct device *in);
void device_timestep(struct device *in);

FILE *fopena(char *path,char *name,const char *mode);
char * get_arg_plusone( char* in[],int count,char * find);
int get_arg_plusone_pos( char* in[],int count,char * find);
int scanarg( char* in[],int count,char * find);
void fit_tpv(char *file, double start, double stop,double a,double b,double c);
double get_I_recomb(struct device *in);
double wire_get_voltage_at_end();
void wire_set_dt(double dt_step);
double wire_getvoltage();
void wire_addtime();
void wire_init(double in,double timestep);
double wire_sim(double Vleft);
double newton_sim_tpv(struct device *in);
void sim_tpv(struct device *in);
int wire();
int hashget(double *x,int N,double find);

void dump_for_plot(struct device *in);
void md5write();
int md5check();
int run_simulation_remove(char *name);
void edit_file_int(char *in_name,char *front,int line,int value);
double get_taue0(int type);
double get_taue1(int type);
double get_tauh0(int type);
double get_tauh1(int type);
double sim_externalv(struct device *in,double wantedv);
double sim_i(struct device *in,double wantedi);
void edit_file(char *in_name,char *front,int line,double value);
void dump_qe();
void dump_talbe();

double get_J_left(struct device *in);
double get_J_right(struct device *in);
double get_J_recom(struct device *in);
int fit_now();
int run_simulation(char *name);
double read_value(char *file, int skip,int line);
void copy_file(char *out_name, char *in_name,char *front,int line,double value);



double get_dark_time_error();


double dos_get_Ec_edge(int i);
double dos_get_Ev_edge(int i);
void draw_gaus(struct device *in);

void sim_jv(struct device *in);

double carrier_count_get_rn(struct device *in);
double carrier_count_get_rp(struct device *in);

double carrier_count_get_n(struct device *in);
double carrier_count_get_p(struct device *in);
void carrier_count_reset(struct device *in);
void carrier_count_add(struct device *in);
double get_extracted_n(struct device *in);
double get_extracted_p(struct device *in);
void device_timestep(struct device *in);
void solve_all(struct device *in);

double get_equiv_V(struct device *in);
double get_equiv_J(struct device *in);
void solver_free_memory(struct device *in);
void save_config(char *simfile,struct device *in);
void calculate_kd(struct device *in);
void update_heat(struct device *in);
void plot_now_excite(struct device *in);
double get_excite_error(struct device *in,double *b);
int solve_excite(struct device *in);
void black();
void device_update_photon_dist(struct device *in);
double B(double x);
double dB(double x);

double get_cur_error(struct device *in);
void update_solver_vars(struct device *in,int clamp);
void do_backtrack(struct device *in);
void fill_matrix(struct device *in);
void plot_open(struct device *in);
void solver_precon(int col,int nz,int *Ti,int *Tj, double *Tx,double *b);
void solver_dump_matrix(int col,int nz,int *Ti,int *Tj, double *Tx,double *b);
void device_init(struct device *in);
void load_config(char *simfile,struct device *in,int gui);
double get_I(struct device *in);
void update(struct device *cell);
int solve_cur(struct device *in);
int solve_cur_thermal(struct device *in,int thermal);
double mod(double in);
int solve_pos(struct device *in);
void get_initial(struct device *in);
double get_mat_param_lin(struct layer* in,int param,double x);
double get_mat_param(struct layer* in,int param);
double get_mat_param_poly(struct layer* in,int param,double x);
void mat_load_poly(FILE *file,struct param* in);
void load_mat_file(char *name,struct layer* in);
int solver(int col,int nz,int *Ti,int *Tj, double *Tx,double *b);
void solver_dump_matrix(int col,int nz,int *Ti,int *Tj, double *Tx,double *b);
void solver_realloc(struct device *in);
void printdebug(char *text);

void backtrak_save(struct device *in);
void backtrack_restore(struct device *in);
void plot_now(struct device *in,char *name);
void plot_close(struct device *in);
void plot_replot(struct device *in);
FILE* start_gnuplot ();
double get_J(struct device *in);
void dump_thermal(struct device *in);

double get_thermal_error(struct device *in,double *b);
int solve_thermal(struct device *in);
void device_get_memory(struct device *in);
void device_free(struct device *in);
void sim_fish(struct device *in);

//DOS model

void gen_dos_maxwell_boltzmann();
void gen_dos_fd();

void gen_dos_fd_gaus_fd();
void gen_dos_maxwell_boltzmann();


//end of DOS model
double get_charge(struct device *in);
int get_file_len(char *file_name);
double get_mat_param_mobility(struct layer* in,int param,int n);
double get_mat_param_xy(struct layer* in,int param,double x);
void mat_free(struct layer* in);
void dos_set_cap_rate(double capeff);
double get_extracted_np(struct device *in);
double get_extracted_k(struct device *in);

void light_field(struct device *in,char *name,double Plaser);
void rbi_test(struct device *in);
void dirty_fit();
double get_mat_param_dmobility(struct layer* in,int param,double n);

double get_bi_error(int ymeshpoints,double *b);
void solve_bi();

double get_avg_mue(struct device *in);
double get_avg_muh(struct device *in);


double get_free_n_charge(struct device *in);
double get_free_p_charge(struct device *in);
double get_total_n_trapped_charge(struct device *in);
double get_total_p_trapped_charge(struct device *in);
void solver_free();

double sim_voc(struct device *in);
void newton_sim_tpv_fast(struct device *in);
void set_ntricks_fast(int val);
#endif

