// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>

double sim_tpc_step(struct device *in)
{
double clamp=0.1;
double step=0.01;
double e0;
double e1;
double i0;
double i1;
double deriv;
double C=in->area*epsilon0*in->epsilonr[0]/in->ylen;
//double R1=in->Rcontact+in->Rshort;
//double R2=in->Rshunt;
solve_all(in);
i0=get_I(in);
//getchar();
double igen=(i0+in->Vapplied/in->Rshunt+C*(in->Vapplied-in->Vapplied_last)/in->dt);

e0=fabs(igen*in->Rcontact+in->Vapplied-in->externalv);
in->Vapplied+=step;
solve_all(in);
i1=get_I(in);
igen=(i1+in->Vapplied/in->Rshunt+C*(in->Vapplied-in->Vapplied_last)/in->dt);
e1=fabs(igen*in->Rcontact+in->Vapplied-in->externalv);
printf("%le\n",e1);
deriv=(e1-e0)/step;
step=-e1/deriv;
//step=step/(1.0+fabs(step/clamp));
in->Vapplied+=step;
int count=0;
int max=1000;
do
{
e0=e1;
solve_all(in);
igen=(i1+in->Vapplied/in->Rshunt+C*(in->Vapplied-in->Vapplied_last)/in->dt);
e1=fabs(igen*in->Rcontact+in->Vapplied-in->externalv);
printf("%le %le %le %le\n",igen*in->Rcontact,in->Vapplied,-in->externalv,e1);
deriv=(e1-e0)/step;
step=-e1/deriv;
//double clamp=0.01;
//if (e1<clamp) clamp=e1/100.0;
//step=step/(1.0+fabs(step/clamp));
step=step/(1.0+fabs(step/clamp));
in->Vapplied+=step;
//printf("ce error = %e %e\n",in->Vapplied,e1);
if (count>max) break;
count++;
}while(e1>1e-5);
//printf("result = %e %e %e %e %e\n",i1,in->Vapplied/Rshort,C*(in->Vapplied-in->Vapplied_last)/in->dt,i1+in->Vapplied/Rshort+C*(in->Vapplied-in->Vapplied_last)/in->dt);
double ret=in->Vapplied-C*(i1-in->Ilast)/in->dt;
return ret;
}


void sim_tpc(struct device *in)
{
int i;
in->dt=in->tpcdtstart;
//last_light=-1000.0;

carrier_count_reset(in);

in->go_time=FALSE;

struct istruct tpcfitout;
inter_init(&tpcfitout);

struct istruct gexout;
inter_init(&gexout);

struct istruct Rnout;
inter_init(&Rnout);

struct istruct Rpout;
inter_init(&Rpout);

struct istruct Rnpout;
inter_init(&Rnpout);

struct istruct ntrapout;
inter_init(&ntrapout);

struct istruct ptrapout;
inter_init(&ptrapout);

struct istruct tpcextracted;
inter_init(&tpcextracted);

struct istruct tpcextracted2;
inter_init(&tpcextracted2);

printf("Running tpc. %e\n",in->externalv);
double V=in->Vstart;
int ittr=0;
//double total_steps=0;
int step=0;

light_field(in,in->spectra,0.0);

sim_externalv(in,in->externalv);
time_init(in);
device_timestep(in);

in->go_time=TRUE;
V=in->Vapplied;

in->time=-in->pulsestart*in->dt;
int dump_count=0;
int tpc_running=FALSE;
double np_out_tot=0.0;
double Psun_orig=in->Psun;
//FILE *srhout=fopen("./srhrates.dat","w");
struct map frame;
frame_init(&frame,20,100,in->ymesh[0],in->ymesh[in->ymeshpoints-1],-6.0,-4.0);

do
{
if (step==in->pulsestart) carrier_count_reset(in);

	if (step<in->pulsestart)
	{
		light_update(in,0.0);
		reset_npinit(in);
	}else
	if ((step>=in->pulsestart)&&(in->time<(in->tpclaseroff)))
	{
		light_update(in,in->tpclaserpower);
		//printf("on\n");
		//if (in->stop_start==TRUE) getchar();
	}else
	{
		if (tpc_running==FALSE)
		{
			np_out_tot=0.0;
			tpc_running=TRUE;
		}
		light_update(in,0.0);
		//printf("off\n");
		//if (in->stop_start==TRUE) getchar();
	}

	in->Vapplied=V;

	frame_reset(&frame);
	frame_build(&frame,in);
	frame_dump(&frame,in);

	solve_all(in);

	//sim_tpc_step(in);
	#ifndef _fast_
	printf("tpc time=%e step=%d %.1e ",in->time,step,in->last_error);
	char out[100];
	sprintf(out,"V=%lf current = %e mA (%e A/m^2)\n",V,get_I(in)/1e-3,get_J(in));

	printdebug(out);
	#endif


	inter_append(&tpcfitout,in->time,get_equiv_I(in));
	inter_append(&gexout,in->time,in->Gex[0]);

	inter_append(&Rnout,in->time,get_avg_recom_n(in));
	inter_append(&Rpout,in->time,get_avg_recom_p(in));

	inter_append(&Rnpout,in->time,get_avg_recom_n(in)-get_avg_recom_p(in));

	inter_append(&ntrapout,in->time,get_n_trapped_charge(in));

	inter_append(&ptrapout,in->time,get_p_trapped_charge(in));

	double a=0.0;
	double b=0.0;
	double c=0.0;
	double d=0.0;

	for (i=0;i<in->ymeshpoints;i++)
	{
		a+=in->nrelax[i];
		b+=in->ntrap_to_p[i];
		c+=in->prelax[i];
		d+=in->ptrap_to_n[i];
	}
	a/=in->ymeshpoints;
	b/=in->ymeshpoints;
	c/=in->ymeshpoints;
	d/=in->ymeshpoints;

	//fprintf(srhout,"%le %le %le %le %le %le %le\n",in->time,a,b,c,d,get_free_n_charge(in),get_free_p_charge(in));

	ittr++;

	plot_now(in,"plot");

	if (in->time>in->tpclaseroff) in->dt*=in->tpcdtmull;

	if (tpc_running==TRUE)
	{
		//np_out_tot+=;
		inter_append(&tpcextracted,in->dt,2.0*get_J(in)/in->ylen/q);
		inter_append(&tpcextracted2,in->time,2.0*get_J(in)/in->ylen/q);
	}
	device_timestep(in);
	step++;

	if (in->time>in->tpcstop) break;

	if (in->stop==TRUE)
	{
		break;

	}


}while(1);

//fclose(srhout);

in->go_time=FALSE;

char outpath[200];

#ifndef mindump
inter_save_a(&tpcextracted2,in->outputpath,"tpcextracted2.dat");
#endif
inter_sub(&tpcextracted,tpcextracted.data[tpcextracted.len-1]);
double tot=0.0;
for (i=0;i<tpcextracted.len;i++)
{
tpcextracted.x[i]=tpcextracted2.x[i];
tot+=tpcextracted.data[i]*tpcextracted.x[i];
tpcextracted.data[i]=tot;
}

inter_sub(&tpcfitout,tpcfitout.data[tpcfitout.len-1]);

sprintf(outpath,"%s%s",in->outputpath,"tpcfit_nomod.dat");
inter_rescale(&tpcfitout,1.0, -1.0);
#ifndef mindump
inter_save(&tpcfitout,outpath);
#endif

inter_mod(&tpcfitout);

//if (tpcfitout.data[tpcfitout.len/2]<0.0) inter_rescale(&tpcfitout,1.0, -1.0);
sprintf(outpath,"%s%s",in->outputpath,"tpcfit.dat");
inter_save(&tpcfitout,outpath);

if (in->dumpitdos==TRUE)
{
	inter_chop(&tpcfitout,1e-9, 1.0);

	for (i=0;i<tpcfitout.len;i++)
	{
	tpcfitout.data[i]*=tpcfitout.x[i];
	tpcfitout.x[i]=(25e-3*log(2e7*tpcfitout.x[i]));
	}

	//inter_chop(&tpcfitout,1e-9, 1.0);

	double val=inter_get_hard(&tpcfitout,0.0);

	inter_rescale(&tpcfitout,1.0, 1.0/val);

	sprintf(outpath,"%s%s",in->outputpath,"tpcitdos.dat");
	inter_save(&tpcfitout,outpath);

}



inter_free(&tpcfitout);

#ifndef mindump
sprintf(outpath,"%s%s",in->outputpath,"gex.dat");
inter_save(&gexout,outpath);
inter_free(&gexout);


sprintf(outpath,"%s%s",in->outputpath,"Rntpc.dat");
inter_save(&Rnout,outpath);
inter_free(&Rnout);

sprintf(outpath,"%s%s",in->outputpath,"Rptpc.dat");
inter_save(&Rpout,outpath);
inter_free(&Rpout);

sprintf(outpath,"%s%s",in->outputpath,"Rnptpc.dat");
inter_save(&Rnpout,outpath);
inter_free(&Rnpout);

sprintf(outpath,"%s%s",in->outputpath,"ntraptpc.dat");
inter_save(&ntrapout,outpath);
inter_free(&ntrapout);

sprintf(outpath,"%s%s",in->outputpath,"ptraptpc.dat");
inter_save(&ptrapout,outpath);
inter_free(&ptrapout);

inter_save_a(&tpcextracted,in->outputpath,"tpcextracted.dat");
#endif

FILE *info=fopen("./tpcinfo.dat","w");
fprintf(info,"%le %le %le\n",Psun_orig,in->externalv,inter_get_hard(&tpcextracted,10e-3));
fclose(info);

inter_free(&tpcextracted);
inter_free(&tpcextracted2);

frame_free(&frame);
}
//</clean>
